// Verified to compile correctly with avr-gcc 5.4.0

// The device must run at factory fuses: 9.6 MHz frequency with CKDIV8 enabled,
// therefore the CPU clock must run at 1.2 MHz.
//
// We reprogram the fuses to use the 128 kHz clock without CKDIV8.
//#define F_CPU 128000UL // this line must be before #include <util/delay.h>!
#define F_CPU 1200000UL // this line must be before #include <util/delay.h>!

// We assume 20% accuracy of the CPU frequency, so the Timer/Counter accuracy
// is also assumed to be equal to 20%.
#define F_CPU_ACCURACY_PERCENT 20

#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdbool.h>


// ADC multiplexer constants, as defined by the ATTiny13A docs (ADMUX
// register), to select the ADC listening pin
#define ADC_ON_PB2 1
#define ADC_ON_PB3 3
#define ADC_ON_PB4 2
#define ADC_ON_PB5 0

// PB0: enable DC-DC converter
// PB1: unconnected
// PB2: voltage measurement
// PB3: current-measuring resistor
// PB4: modulation
// PB5: reset/reserved

// Signalling MODULATION on PB2
#define BIT_MODULATION 4
#define BIT_POWER_GOOD 0

// Launch the ADC once and wait for it to finish (synchronous).
// The 10-bit reading will be stored in the 16-bit register ADCW.
#define adc_fire_once(){ \
	ADCSRA |= (1 << ADSC); \
	loop_until_bit_is_clear(ADCSRA, ADSC); \
}


#include "tinyqirx.c"

void __attribute__ ((noinline)) qi_send_bit(qi_bit_t bit)
{
	if (bit != QI_500USEC_PAUSE) PORTB ^= (1 << BIT_MODULATION);
	_delay_us(460);
	if (bit == QI_BIT_ONE) PORTB ^= (1 << BIT_MODULATION);
	_delay_us(460);
}


uint8_t __attribute__ ((noinline)) qi_measure_voltage(){
	// - select the source of the Analog signal;
	// - left-adjust the result
	// - set Internal 1.1 V voltage reference.
	ADMUX = ADC_ON_PB2 | (1 << REFS0) | (1 << ADLAR);

	// Also let it run once, to initialize.
	adc_fire_once();
	return ADCH;
}

uint8_t __attribute__ ((noinline)) qi_measure_current(){
	// - select the source of the Analog signal;
	// - set Internal 1.1 V voltage reference.
	ADMUX = ADC_ON_PB3 | (1 << REFS0);

	// Also let it run once, to initialize.
	adc_fire_once();
	uint16_t result = ADCW;
	return (uint8_t) result;
}

int main(){
	// Start up the ADC:
	{
		// - select the source of the Analog signal;
		// - left-adjust the result
		// - set Internal 1.1 V voltage reference.
		ADMUX = ADC_ON_PB2 | (1 << REFS0) | (1 << ADLAR);

		// - enable the Analog-Digital converter in the Single Conversion mode; 
		// - set the frequency to 1/16 of the CPU frequency (< 200 kHz) to
		// ensure 10 bit conversion.
		ADCSRA = (1 << ADEN) | 3;

		// Also let it run once, to initialize.
		adc_fire_once();
	}


	// Enable Write on the MODULATION pin and DC-DC enabled pin
	DDRB |= (1 << BIT_MODULATION) | (1 << BIT_POWER_GOOD);
	// Set them to low
	PORTB &= ~((1 << BIT_MODULATION) | (1 << BIT_POWER_GOOD));

	qi_signal_power_good(false);

	// We are ready to go, set Global Enable interrupts
	sei();

	qi_connect_to_transmitter();
	while (1){
		if(qi_periodic_cb()) {
			_delay_ms(300);
			// Send POWER GOOD
			PORTB |= (1 << BIT_POWER_GOOD);
		} else {
			// Turn off POWER GOOD
			PORTB &= ~(1 << BIT_POWER_GOOD);
		}
	}

	return 0;
}
