#include <stdint.h>
#include <stdbool.h>


#include "tinyqirx_utils.c"
#include "tinyqirx_core.c"



uint8_t qi_measure_voltage();
uint8_t qi_measure_current();

void qi_connect_to_transmitter(){
	uint8_t voltage = qi_measure_voltage();
	qi_send_signal_strength_packet(voltage, _TINYQIRX_MAXIMUM_EXPECTED_VOLTAGE_BITS);
	qi_send_identification(_TINYQIRX_QI_MANUFACTURER_CODE, _TINYQIRX_QI_DEVICE_IDENTIFIER);
	qi_send_configuration(_TINYQIRX_REFERENCE_POWER_BITS, _TINYQIRX_WINDOW_SIZE_BITS, _TINYQIRX_WINDOW_OFFSET_BITS);
}

#define qi_end_power_receive(error_code) qi_send_end_of_power_transfer_packet(error_code)

uint8_t qi_try_restart(){
#ifdef _TINYQIRX_TEMPERATURE_CONTROL
	if (qi_is_battery_overheating()) {
		return QI_EPT_OVERTEMPERATURE;
	}
#endif
	uint8_t voltage_bits = qi_measure_voltage();
	if (voltage_bits > _TINYQIRX_MAXIMUM_EXPECTED_VOLTAGE_BITS) {
		return QI_EPT_OVERVOLTAGE;
	}
	uint8_t current_bits = qi_measure_current();
	if (current_bits > _TINYQIRX_MAXIMUM_CURRENT_BITS) {
		return QI_EPT_OVERCURRENT;
	}
#ifdef _TINYQIRX_CUTOFF_CURRENT_BITS
	if (current_bits < _TINYQIRX_CUTOFF_CURRENT_BITS) {
		return QI_EPT_CHARGE_COMPLETE;
	}
#endif
	qi_send_end_of_power_transfer_packet(QI_EPT_RESTART);
	return QI_EPT_RESTART;
}


bool qi_periodic_cb(){
	uint32_t averaged_power_voltAmpereBits = 0;
	for(int8_t sample = 0; sample < _TINYQIRX_POWER_ESTIMATION_SAMPLES; sample++){
#ifdef _TINYQIRX_TEMPERATURE_CONTROL
		if (qi_is_battery_overheating()) {
			qi_send_end_of_power_transfer_packet(QI_EPT_OVERTEMPERATURE);
			return false;
		}
#endif
		uint8_t voltage_bits = qi_measure_voltage();

		if (voltage_bits > _TINYQIRX_MAXIMUM_EXPECTED_VOLTAGE_BITS) {
			qi_send_end_of_power_transfer_packet(QI_EPT_OVERVOLTAGE);
			return false;
		}
		if ((voltage_bits > _TINYQIRX_VOLTAGE_TOLERANCE_RANGE_UPPER_LIMIT_BITS)
		||  (voltage_bits < _TINYQIRX_VOLTAGE_TOLERANCE_RANGE_LOWER_LIMIT_BITS)){
			qi_send_control_error_packet(voltage_bits, _TINYQIRX_SETPOINT_VOLTAGE_BITS);
			// Quit sampling the power and measure the voltage rapidly instead
			// until it converges to the allowed range
			sample = -1;
			continue;
		}
		uint8_t current_bits = qi_measure_current();
		if (current_bits > _TINYQIRX_MAXIMUM_CURRENT_BITS) {
			qi_send_end_of_power_transfer_packet(current_bits);
			return false;
		}
#ifdef _TINYQIRX_CUTOFF_CURRENT_BITS
		if (current_bits < _TINYQIRX_CUTOFF_CURRENT_BITS) {
			qi_signal_power_good(false);
			qi_send_end_of_power_transfer_packet(QI_EPT_CHARGE_COMPLETE);
			return false;
		}
#endif
		averaged_power_voltAmpereBits += voltage_bits * current_bits;
		qi_send_bit(QI_500USEC_PAUSE);
	}
	uint8_t pause;
	for (pause = 0; pause < _TINYQIRX_WINDOW_OFFSET_FILL; pause++){
		qi_send_bit(QI_500USEC_PAUSE);
	}
	if (averaged_power_voltAmpereBits != 0){
#ifdef _TINYQIRX_USE_APPROXIMATE_DIVISION
/* Approximate division that relies on the compiler for performing the division
 * operation and multiplies the result. For uint8_t, approximate division only
 * differs from precise division by +/-1 at most. */
		uint8_t averaged_power_bits = (uint8_t) ((_TINYQIRX_2_TO_24TH_DIVIDED_BY_POWER_CONVERSION_FACTOR * (averaged_power_voltAmpereBits >> _TINYQIRX_POWER_PRECISION_SHIFT)) >> 16);
#else
		uint8_t averaged_power_bits = (uint8_t) (averaged_power_voltAmpereBits / _TINYQIRX_POWER_CONVERSION_DIVISOR);
#endif
		qi_send_received_power_packet_bits(averaged_power_bits);
	} else {
		qi_send_received_power_packet_bits(0);
	}
	return true;
}

