#include <stdint.h>

#include "tinyqirx_conf.auto.h"

#include "tinyqirx_utils.c"

#define QI_EPT_GENERIC 0x00
#define QI_EPT_CHARGE_COMPLETE 0x01
#define QI_EPT_INTERNAL_FAULT 0x02
#define QI_EPT_OVERTEMPERATURE 0x03
#define QI_EPT_OVERVOLTAGE 0x04
#define QI_EPT_OVERCURRENT 0x05
#define QI_EPT_BATTERY_FAILURE 0x06
#define QI_EPT_NO_RESPONSE 0x08
#define QI_EPT_ABORTED_NEGOTIATION 0x0A
#define QI_EPT_RESTART 0x0B
#define QI_EPT_REPING 0x0C
#define QI_EPT_RFID 0x0D


typedef enum {
	QI_BIT_ZERO = 0,
	QI_BIT_ONE = 1,
	QI_500USEC_PAUSE = 0xFF
} qi_bit_t;

void qi_send_bit(qi_bit_t bit);

#define qi_send_control_error_packet(actual_voltage, setpoint_voltage) { \
	uint8_t ce_bytes[] = {0x03, ((setpoint_voltage) - (actual_voltage))}; \
	qi_send_packet(ce_bytes, 2); \
}


#ifdef _TINYQIRX_USE_APPROXIMATE_DIVISION
#define qi_send_signal_strength_packet(actual_voltage, maximum_expected_voltage) { \
	uint8_t sig_bytes[2]; \
	sig_bytes[0] = 0x01; /* SIG packet identifier */ \
	if ((actual_voltage) >= (maximum_expected_voltage)) { \
		sig_bytes[1] = 0xFF; \
	} else { \
	/* Approximate division that relies on the compiler for performing the division */ \
	/* operation and multiplies the result. For uint8_t, approximate division only */ \
	/* differs from precise division by +/-1 at most. */ \
		sig_bytes[1] = ((uint8_t) (((_TINYQIRX_2_TO_24TH_DIVIDED_BY_MAXIMUM_EXPECTED_VOLTAGE_BITS) * ((uint16_t) actual_voltage)) >> 16)); \
	} \
	qi_send_packet(sig_bytes, 2); \
}
#else
#define qi_send_signal_strength_packet(actual_voltage, maximum_expected_voltage) { \
	uint8_t sig_bytes[2]; \
	sig_bytes[0] = 0x01; /* SIG packet identifier */ \
	if ((actual_voltage) >= (maximum_expected_voltage)) { \
		sig_bytes[1] = 0xFF; \
	} else { \
		sig_bytes[1] = ((((uint16_t) (actual_voltage)) << 8) / (maximum_expected_voltage)); \
	} \
	qi_send_packet(sig_bytes, 2); \
}
#endif

#define qi_send_received_power_packet_bits(actual_power_bits) { \
	uint8_t rp8_bytes[2] = {0x04, actual_power_bits}; \
	qi_send_packet(rp8_bytes, 2); \
}

#define qi_send_end_of_power_transfer_packet(error_code) { \
	uint8_t ept_bytes[2]; \
	ept_bytes[0] = 0x02; /* EPT packet identifier */ \
	ept_bytes[1] = (error_code); \
	qi_send_packet(ept_bytes, 2); \
}

#define qi_send_identification(manufacturer_code, device_identifier) { \
	uint8_t id_bytes[8]; \
	const uint8_t MAJOR_VERSION = 1; \
	const uint8_t MINOR_VERSION = 0; \
	id_bytes[0] = 0x71; /* ID packet identifier */ \
	id_bytes[1] = MINOR_VERSION << 4 | MAJOR_VERSION; \
	id_bytes[2] = (uint8_t) (manufacturer_code); \
	id_bytes[3] = (uint8_t) ((manufacturer_code) >> 8); \
	id_bytes[4] = (uint8_t) ((device_identifier) >> 24) & 0x7F; \
	id_bytes[5] = (uint8_t) ((device_identifier) >> 16); \
	id_bytes[6] = (uint8_t) ((device_identifier) >> 8); \
	id_bytes[7] = (uint8_t) (device_identifier); \
	qi_send_packet(id_bytes, 8); \
}

#define qi_send_configuration(reference_power, window_size, window_offset) { \
	uint8_t cfg_bytes[6]; \
	cfg_bytes[0] = 0x51; /* CFG packet identifier */ \
	cfg_bytes[1] = (reference_power) & 0x3f; \
	cfg_bytes[2] = 0x00; /* Reserved */ \
	cfg_bytes[3] = 0x00; /* Extended protocol functionality */ \
	cfg_bytes[4] = ((window_size) << 4) | ((window_offset) & 0x0F); \
	cfg_bytes[5] = 0x00; /* Extended protocol functionality */ \
	qi_send_packet(cfg_bytes, 6); \
}

void qi_send_byte(uint8_t byte){
	qi_send_bit(QI_BIT_ZERO);
	uint8_t parity = 1;
	uint8_t i = 0;
	for(i = 0; i < 8; i++){
		qi_send_bit((qi_bit_t) (byte & 1));
		parity ^= (byte & 1);
		byte >>= 1;
	}
	qi_send_bit((qi_bit_t) parity);
	qi_send_bit(QI_BIT_ONE);
}

void qi_send_packet(uint8_t const * const bytes, uint8_t const length){
	uint8_t checksum = 0x00;
	uint8_t i;
	for(i = _TINYQIRX_PREAMBLE_LENGTH_BITS; i > 0; i--){
		qi_send_bit(QI_BIT_ONE);
	}
	i = 0;
	do {
		uint8_t byte = bytes[i];
		qi_send_byte(byte);
		checksum ^= byte;
		i++;
	} while (i < length); 
	qi_send_byte(checksum);
	// This is to switch off the modulation (the last edge) + a portion of the window
	qi_send_bit(QI_BIT_ZERO);
	// If the packet had an odd length, we do it twice to ensure a low level
	if ((length & 1) != 0){
		qi_send_bit(QI_BIT_ZERO);
	} else {
		qi_send_bit(QI_500USEC_PAUSE);
	}
	for(i = 0; i < _TINYQIRX_SILENT_WINDOW_BITS - 2; i++){
		qi_send_bit(QI_500USEC_PAUSE);
	}
}

